using System;
using System.IO;
using System.Security.Cryptography;
using Osci.Common;
using Osci.Cryptographic;
using Osci.Extensions;

namespace Osci.Helper
{
    /// <summary> Diese Erweiterung der Klasse java.io.FilterInputStream führt eine 2-Key-TripleDES-
    /// Ver-/Entschlüsselung der gelesenen Daten im CBC-Mode durch. Der Initialisierungsvektor
    /// wird gemäß der XML-Encryption-Spezifikation den Inhaltsdaten hinzugefügt bzw.
    /// entnommen.
    /// 
    /// <p>Copyright © 2018 Freie Hansestadt Bremen</p>
    /// <p>Erstellt von Governikus GmbH &amp; Co. KG</p>
    /// <p>Diese Bibliothek kann von jedermann nach Maßgabe der Bremer Lizenz für
    /// freie Softwarebibliotheken genutzt werden.</p><p>Die Lizenzbestimmungen können
    /// unter der URL www.osci.de abgerufen oder bei der
    /// Koordinierungsstelle für IT-Standards (KoSIT), Schillerstr. 1,
    /// 28195 Bremen schriftlich angefordert werden.</p>
    /// 
    /// <p>Author: P. Ricklefs, N. Büngener</p> 
    /// <p>Version: 1.8</p>
    /// </summary>
    /// <seealso cref="SymCipherOutputStream">
    /// </seealso>
    public class SymCipherInputStream
        : InputStream
    {
        private static readonly Log _log = LogFactory.GetLog(typeof(SymCipherInputStream));
        private string _cipherName;
        private int _count;
        private readonly bool _encrypt;
        private readonly CryptoStream _baseStream;
        private readonly SecretKey _secretKey;

        public SymCipherInputStream(Stream stream, SecretKey secretKey, bool encrypt = false)
        {
            _encrypt = encrypt;
            _secretKey = secretKey;

            if (stream == null)
            {
                throw new ArgumentNullException("stream");
            }
            if (secretKey == null)
            {
                throw new ArgumentNullException("secretKey", "Es wurde kein symmetrischer Schlüssel übergeben.");
            }

            _log.Trace("CipherAlgorithm: " + secretKey.AlgorithmType);

            ICryptoTransform cryptoTransform;
            if (encrypt)
            {
                secretKey.GenerateIV();
                cryptoTransform = secretKey.CreateEncryptor();
            }
            else
            {
                byte[] iv = new byte[secretKey.IVSize];
                stream.Read(iv, 0, iv.Length);
                _secretKey.IV = iv;
                cryptoTransform = secretKey.CreateDecryptor();
            }

            _baseStream = new CryptoStream(stream, cryptoTransform, CryptoStreamMode.Read);
        }

        public override int Read(byte[] buffer, int off, int len)
        {
            byte[] iv = _secretKey.IV;

            if (_encrypt && _count < iv.Length)
            {
                if (len < iv.Length - _count)
                {
                    Array.Copy(iv, _count, buffer, off, len);
                    _count += len;
                    return len;
                }
                else
                {
                    Array.Copy(iv, _count, buffer, off, iv.Length - _count);
                    int ret = _baseStream.CopyTo(buffer, off + iv.Length - _count, len - iv.Length + _count);
                    ret += iv.Length - _count;
                    _count = iv.Length;
                    return ret;
                }
            }
            else
            {
                try
                {
                    return _baseStream.Read(buffer, off, len);
                }
                catch (CryptographicException)
                {
                    // Prevent padding oracle attack: Do not expose cryptographic exceptions
                    return 0;
                }
            }
        }
    }
}