using System;
using System.IO;
using System.Security.Cryptography;
using Osci.Common;
using Osci.Cryptographic;

namespace Osci.Helper
{
    /// <summary> Diese Erweiterung der Klasse java.io.FilterOutputStream führt eine 2-Key-TripleDES-
    /// Ver-/Entschlüsselung der geschriebenen Daten im CBC-Mode durch. Der Initialisierungsvektor
    /// wird gemäß der XML-Encryption-Spezifikation den Inhaltsdaten hinzugefügt bzw.
    /// entnommen.
    /// 
    /// <p>Copyright © 2018 Freie Hansestadt Bremen</p>
    /// <p>Erstellt von Governikus GmbH &amp; Co. KG</p>
    /// <p>Diese Bibliothek kann von jedermann nach Maßgabe der Bremer Lizenz für
    /// freie Softwarebibliotheken genutzt werden.</p><p>Die Lizenzbestimmungen können
    /// unter der URL www.osci.de abgerufen oder bei der
    /// Koordinierungsstelle für IT-Standards (KoSIT), Schillerstr. 1,
    /// 28195 Bremen schriftlich angefordert werden.</p>
    /// 
    /// <p>Author: P. Ricklefs, N. Büngener</p> 
    /// <p>Version: 1.8</p>
    /// </summary>
    /// <seealso cref="SymCipherInputStream">
    /// </seealso>
    public class SymCipherOutputStream
        : OutputStream
    {
        private int _index;
        private readonly SecretKey _secretKey;
        private readonly bool _encrypt;
        private sbyte[] _tmp;

        private CryptoStream _baseStream;
        private Stream _outStream;

        public SymCipherOutputStream(Stream outStream, SecretKey secretKey, byte[] initializationVector)
            : this(outStream, secretKey, true, initializationVector)
        {
        }

        public SymCipherOutputStream(Stream outStream, SecretKey secretKey, bool encrypt)
            : this(outStream, secretKey, encrypt, null)
        {
        }

        internal SymCipherOutputStream(Stream outStream, SecretKey secretKey, bool encrypt, byte[] initializationVector)
        {
            _secretKey = secretKey;
            _encrypt = encrypt;
            if (encrypt)
            {
                if (secretKey == null)
                {
                    secretKey = new SecretKey();
                }

                if (initializationVector != null)
                {
                    secretKey.IV = initializationVector;
                }

                outStream.Write(secretKey.IV, 0, secretKey.IV.Length);
                _baseStream = new CryptoStream(outStream, secretKey.CreateEncryptor(), CryptoStreamMode.Write);
            }
            else
            {
                secretKey.IV = new byte[secretKey.IVSize];
            }
        }

        public override void Write(int b)
        {
            Write(new[] { (byte)b }, 0, 1);
        }

        public override void Write(byte[] data, int offset, int length)
        {
            if ((offset | length | (data.Length - (length + offset)) | (offset + length)) < 0)
            {
                throw new IndexOutOfRangeException();
            }

            if (!_encrypt && _index < _secretKey.IV.Length)
            {
                if (length < (_secretKey.IV.Length - _index))
                {
                    for (int i = 0; i < length; i++)
                    {
                        _secretKey.IV[_index++] = data[offset + i];
                    }
                }
                else
                {
                    int i;

                    for (i = 0; i < _secretKey.IV.Length - _index; i++)
                    {
                        _secretKey.IV[i + _index] = data[offset + i];
                    }

                    _index += i;

                    _baseStream = new CryptoStream(_outStream, _secretKey.CreateDecryptor(), CryptoStreamMode.Write);
                    _baseStream.Write(data, offset + i, length - i);
                }
            }
            else
            {
                _baseStream.Write(data, offset, length);
            }
        }

        public override void Close()
        {
            _baseStream.FlushFinalBlock();
        }
    }
}