using System.Collections;
using System.IO;
using System.Security.Cryptography;
using Osci.Common;
using Osci.Encryption;
using Osci.Extensions;
using Osci.Helper;
using Osci.Roles;

namespace Osci.MessageParts
{
    /// <summary> Diese Klasse bildet eine die Grundlage für XML-Signature Signaturen der Bibliothek.
    /// Sie wird von Anwendungen nie direkt benötigt.
    /// 
    /// <p>Copyright © 2018 Freie Hansestadt Bremen</p>
    /// <p>Erstellt von Governikus GmbH &amp; Co. KG</p>
    /// <p>Diese Bibliothek kann von jedermann nach Maßgabe der Bremer Lizenz für
    /// freie Softwarebibliotheken genutzt werden.</p><p>Die Lizenzbestimmungen können
    /// unter der URL www.osci.de abgerufen oder bei der
    /// Koordinierungsstelle für IT-Standards (KoSIT), Schillerstr. 1,
    /// 28195 Bremen schriftlich angefordert werden.</p>
    /// 
    /// <p>Author: P. Ricklefs, N. Büngener</p> 
    /// <p>Version: 1.8</p>
    /// </summary>
    public class OsciSignature
        : MessagePart
    {
        public byte[] SignedInfo
        {
            get; set;
        }

        public ArrayList Refs
        {
            get; set;
        }

        public byte[] SignatureValue
        {
            get; set;
        }

        public string SignerId
        {
            get; set;
        }

        public string SignatureAlgorithm
        {
            get; set;
        }

        public Role Signer
        {
            get; set;
        }

        internal string EnclosingElement
        {
            get; set;
        }

        internal string SigningTime
        {
            get; set;
        }

        internal string SigningPropsId
        {
            get; set;
        }

        internal string SigningProperties
        {
            get; set;
        }

        internal Hashtable RefsHash
        {
            get; set;
        }

        internal Hashtable RefsDigestMethods
        {
            get; set;
        }

        internal static int Cnt = 0;
        private readonly static Log _log = LogFactory.GetLog(typeof(OsciSignature));


        internal OsciSignature()
        {
            Refs = new ArrayList();
            SignatureAlgorithm = DialogHandler.SignatureAlgorithm;
        }

        internal OsciSignature(string enclosingElement)
            : this()
        {
            EnclosingElement = enclosingElement;
        }

        internal void AddSignatureTime(string time, string id, string digestAlgorithm)
        {
            SigningPropsId = id;
            SigningTime = time;
            string nsHead = "xmlns:ds=\"http://www.w3.org/2000/09/xmldsig#\" xmlns:osci=\"http://www.osci.de/2002/04/osci\" xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\" xmlns:xades=\"http://uri.etsi.org/01903/v1.3.2#\" xmlns:xenc=\"http://www.w3.org/2001/04/xmlenc#\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"";
            SigningProperties = "<xades:SignedProperties " + nsHead
                + " Id=\"" + SigningPropsId
                + "\"><xades:SignedSignatureProperties><xades:SigningTime>"
                + SigningTime
                + "</xades:SigningTime></xades:SignedSignatureProperties></xades:SignedProperties>";

            OsciSignatureReference osr = new OsciSignatureReference("#" + id);

            HashAlgorithm msgDigest = Crypto.CreateMessageDigest(digestAlgorithm);
            osr.digestValue = msgDigest.ComputeHash(SigningProperties.ToByteArray());
            osr.digestMethodAlgorithm = digestAlgorithm;
            osr.SetNamespacePrefixes(SoapNsPrefix, OsciNsPrefix, DsNsPrefix,
                              XencNsPrefix, XsiNsPrefix);
            Refs.Add(osr);
        }

        internal OsciSignatureReference[] GetReferences()
        {
            return (OsciSignatureReference[])SupportClass.CollectionSupport.ToArray(Refs, new OsciSignatureReference[Refs.Count]);
        }

        public void Sign(Role signer)
        {
            Signer = signer;
            SignatureAlgorithm = signer.SignatureAlgorithm;
            if (SignedInfo == null)
            {
                CreateSignedInfo();
            }
            SignatureValue = signer.Signer.Sign(SignedInfo, SignatureAlgorithm);
            SignerId = "#" + signer.SignatureCertificateId;
        }

        /// <summary>Liefert den zu dieser Signatur gehörende Signaturzeitpunkt (ISO-8601-DateTime Format).
        /// </summary>
        /// <returns>Signaturzeitpunkt
        /// </returns>
        public string GetSigningTime()
        {
            return SigningTime;
        }

        public Hashtable GetDigests()
        {
            if (RefsHash == null)
            {
                RefsHash = new Hashtable();
                for (int i = 0; i < Refs.Count; i++)
                {
                    RefsHash.Put(((OsciSignatureReference)Refs[i]).RefId, ((OsciSignatureReference)Refs[i]).digestValue);
                }
            }
            return RefsHash;
        }

        public Hashtable GetDigestMethods()
        {
            if (RefsDigestMethods == null)
            {
                RefsDigestMethods = new Hashtable();

                for (int i = 0; i < Refs.Count; i++)
                {
                    RefsDigestMethods.Add(((OsciSignatureReference)Refs[i]).RefId, ((OsciSignatureReference)Refs[i]).digestMethodAlgorithm);
                }
            }
            return RefsDigestMethods;
        }


        private void CreateSignedInfo()
        {
            using (MemoryStream memoryStream = new MemoryStream())
            {
                memoryStream.Write("<" + DsNsPrefix + ":SignedInfo xmlns:" + DsNsPrefix + "=\"http://www.w3.org/2000/09/xmldsig#\" xmlns:" + OsciNsPrefix + "=\"http://www.osci.de/2002/04/osci\" xmlns:" + SoapNsPrefix + "=\"http://schemas.xmlsoap.org/soap/envelope/\" xmlns:" + XencNsPrefix + "=\"http://www.w3.org/2001/04/xmlenc#\" xmlns:" + XsiNsPrefix + "=\"http://www.w3.org/2001/XMLSchema-instance\">\n");
                memoryStream.Write("<" + DsNsPrefix + ":CanonicalizationMethod Algorithm=\"http://www.w3.org/TR/2001/REC-xml-c14n-20010315\"></" + DsNsPrefix + ":CanonicalizationMethod>\n<" + DsNsPrefix + ":SignatureMethod Algorithm=\"" + SignatureAlgorithm + "\"></" + DsNsPrefix + ":SignatureMethod>\n");

                for (int i = 0; i < Refs.Count; i++)
                {
                    ((OsciSignatureReference)Refs[i]).WriteXml(memoryStream);
                }
                memoryStream.Write("</" + DsNsPrefix + ":SignedInfo>");
                SignedInfo = memoryStream.ToArray();
            }
        }

        public override void WriteXml(Stream stream)
        {
            if (EnclosingElement != null)
            {
                stream.Write(EnclosingElement);
            }
            string prefix = DsNsPrefix + ":";
            int start;
            int stop;
            for (start = 0; start < SignedInfo.Length; start++)
            {
                if (SignedInfo[start] == '<')
                {
                    break;
                }
            }
            byte[] prefixBytes = prefix.ToByteArray();
            int j = 0;
            for (int i = start + 1; i < prefix.Length; i++, j++)
            {
                if (prefixBytes[j] != SignedInfo[i])
                {
                    prefix = "";
                    break;
                }
            }

            if (prefix.Length == 0)
            {
                stream.Write("<Signature " + "xmlns=\"" + MessagePartParser.DsXmlns + "\">");
            }
            else
            {
                stream.Write("<" + prefix + "Signature>");
            }
            //neu in126
            for (start = 0; start < SignedInfo.Length; start++)
            {
                if (SignedInfo[start] == 0x20)
                {
                    break;
                }
            }
            for (stop = start; stop < SignedInfo.Length; stop++)
            {
                if (SignedInfo[stop] == '>')
                {
                    break;
                }
            }
            stream.Write(SignedInfo, 0, start);
            stream.Write(SignedInfo, stop, SignedInfo.Length - stop);
            stream.Write("<" + prefix + "SignatureValue>");
            stream.Write(Helper.Base64.Encode(SignatureValue));
            //id
            stream.Write("</" + prefix + "SignatureValue><" + prefix + "KeyInfo><" + prefix + "RetrievalMethod URI=\"");
            stream.Write(SignerId);

            stream.Write("\"></" + prefix + "RetrievalMethod></" + prefix + "KeyInfo>");
            if (SigningProperties != null)
            {
                stream.Write("<" + DsNsPrefix + ":Object><xades:QualifyingProperties xmlns:xades=\"" + MessagePartParser.XadesXmlns + "\">");
                int nsStart = SigningProperties.IndexOf("<xades:SignedProperties");
                nsStart = SigningProperties.IndexOf(' ', nsStart);
                stream.Write(SigningProperties.Substring(0, nsStart));
                stream.Write(SigningProperties.Substring(SigningProperties.IndexOf(" Id=", nsStart)));
                stream.Write("</xades:QualifyingProperties></" + DsNsPrefix + ":Object>");
            }
            stream.Write("</" + prefix + "Signature>");

            if (EnclosingElement != null)
            {
                stream.Write("</" + EnclosingElement.Substring(1, (EnclosingElement.IndexOf(' ')) - (1)) + ">");
            }
        }
    }
}