using System;
using System.Collections;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Resources;
using Osci.Helper;

namespace Osci.Resources
{
    /// <summary>
    /// Die Unterstützung der mehrsprachigen Ausgaben wurde vereinfacht.
    /// Sprachdateien können nun als einfache Java-properties-Dateien in
    /// das Verzeichnis der DLL osci-bib.dll kopiert werden. Die Dateinamen
    /// müssen aus dem String "Text_" bestehen, dem der CultureInfo.TwoLetterISOLanguageName
    /// sowie die Erweiterung ".properties" angehängt ist, also z.B. "Text_de.properties".
    /// Als default-Datei für unbekannte Sprachen wird die Datei "Text.properties"
    /// geladen.
    /// 
    /// </summary>
    public class ResourceBundle
    {
        public string Path
        {
            get;
        }

        private ResourceManager _rm;
        private readonly Hashtable _props;
        private static readonly Log _log = LogFactory.GetLog(typeof(ResourceBundle));

        private ResourceBundle(string bundleName, CultureInfo cultureInfo)
        {
            _props = new Hashtable();

            try
            {
                FileInfo bundleFile = GetResourceFile(bundleName, cultureInfo);
                Path = bundleFile.FullName;

                foreach (string line in File.ReadAllLines(bundleFile.FullName).Select(_ => _.Trim()).Where(_ => !string.IsNullOrEmpty(_) && !_.StartsWith("#")))
                {
                    string[] keyValue = line.Split(new[] { '=' }, 2);
                    _props.Add(keyValue[0].Trim(), keyValue[1].Trim());
                }
            }
            catch (Exception e)
            {
                _log.Error("ResourceBundle couldn't be initialized properly:", e);
            }
        }

        private FileInfo GetResourceFile(string bundleName, CultureInfo cultureInfo)
        {
            string language = "_" + cultureInfo.TwoLetterISOLanguageName;
            FileInfo assemblyFile = new FileInfo(GetType().Assembly.Location);

            Func<string, FileInfo> buildFileName = l =>
            {
                return new FileInfo(System.IO.Path.Combine(assemblyFile.DirectoryName, string.Format("{0}{1}.properties", bundleName, l)));
            };

            if (buildFileName(language).Exists)
            {
                return buildFileName(language);
            }
            else
            {
                return buildFileName(null);
            }
        }

        public string GetString(string key)
        {
            return _props.ContainsKey(key) ? (string)_props[key] : "";
        }

        public static ResourceBundle GetBundle(string bundleName, CultureInfo cultureInfo)
        {
            return new ResourceBundle(bundleName, cultureInfo);
        }
    }
}
